from django.db import models
from solo.models import SingletonModel


class NotifyConfig(SingletonModel):
    periodic = models.IntegerField(default=60, verbose_name='Периодичность отправки уведомлений (сек.)')

    email_max_day_limit = models.IntegerField(default=240, verbose_name='Дневной лимит отправки писем')
    email_max_hour_limit = models.IntegerField(default=40, verbose_name='Часовой лимит отправки писем')

    sms_url = models.CharField(default='http://sms.ru/sms/send', max_length=255, verbose_name='URL СМС провайдера')
    sms_api_id = models.CharField(default='1234567890', blank=True, max_length=255,
                                  verbose_name='API ID СМС провайдера')
    sms_from = models.CharField(default='', blank=True, max_length=255, verbose_name='Отправитель СМС',
                                help_text='Например, Garpix')

    telegram_api_key = models.CharField(default='000000000:AAAAAAAAAA-AAAAAAAA-_AAAAAAAAAAAAAA', blank=True,
                                        max_length=255, verbose_name='Telegram API Key')
    viber_api_key = models.CharField(default='000000000:AAAAAAAAAA-AAAAAAAA-_AAAAAAAAAAAAAA', blank=True,
                                     max_length=255, verbose_name='Viber API Key')
    viber_bot_name = models.CharField(blank=True, max_length=255, verbose_name='Название viber бота',
                                      default='Viber bot')
    is_email_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку Email')
    is_sms_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку SMS')
    is_push_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку PUSH')
    is_telegram_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку Telegram',
                                              help_text='Внимание! Telegram недоступен на серверах на территории РФ и работать на них не будет!.')
    is_viber_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку Viber')

    viber_success_added_text = models.TextField(blank=True,
                                                default='Ваша учетная запись успешно привязана к боту. Вы будете получать уведомления!',
                                                verbose_name='Viber - Текст успешно добавлен код')
    viber_failed_added_text = models.TextField(blank=True,
                                               default='Ошибка при привязке учетной записи. Пожалуйста, свяжитесь с техподдержкой',
                                               verbose_name='Viber - Текст провал, не добавлен код')

    viber_text_for_new_sub = models.TextField(blank=True,
                                              default='cпасибо за подписку, Введите secret_key чтобы получать сообщения от бота.',
                                              verbose_name='Viber - Текст  для новых подписчиков')

    viber_welcome_text = models.TextField(blank=True,
                                          default='для активации бота нужно отправить любое сообщения',
                                          verbose_name='Viber - Приветственный текст бота')

    class Meta:
        verbose_name = 'Настройка'
        verbose_name_plural = 'Настройки'



from .notify import Notify
from .category import NotifyCategory
from .smtp import SMTPAccount
from .config import NotifyConfig
from .file import NotifyFile
from .fcm import NotifyDevice
from .template import NotifyTemplate
from .user_list import NotifyUserList
from .user_list_participant import NotifyUserListParticipant
from .log import NotifyErrorLog



import json
import re
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Optional
from smtplib import SMTP, SMTP_SSL
import requests
from django.conf import settings
from django.db import models
from django.template import Template, Context
from django.utils.html import format_html
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from viberbot import Api, BotConfiguration
from viberbot.api.messages import TextMessage
from .category import NotifyCategory
from .choices import TYPE, STATE
from .config import NotifyConfig
from .fcm import NotifyDevice
from .file import NotifyFile
from .log import NotifyErrorLog
from .smtp import SMTPAccount
from .template import NotifyTemplate
from ..mixins import UserNotifyMixin


def chunks(s, n):
    """Produce `n`-character chunks from `s`."""
    for start in range(0, len(s), n):
        yield s[start:start + n]


class Notify(UserNotifyMixin):
    """
    Уведомление
    """

    # title = models.CharField(max_length=255, verbose_name='Название для админа')
    subject = models.CharField(max_length=255, default='', blank=True, verbose_name='Тема')
    text = models.TextField(verbose_name='Текст')
    html = models.TextField(verbose_name='HTML', blank=True, default='')

    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL,
                             related_name='notifies', verbose_name='Пользователь (получатель)')
    email = models.EmailField(max_length=255, blank=True, null=True, verbose_name='Email Получателя',
                              help_text='Используется только в случае отсутствия указанного пользователя')
    sender_email = models.EmailField(max_length=255, blank=True, null=True, verbose_name='Email Отправителя')

    type = models.IntegerField(choices=TYPE.CHOICES, verbose_name='Тип')
    state = models.IntegerField(choices=STATE.CHOICES, default=STATE.WAIT, verbose_name='Состояние')
    category = models.ForeignKey(NotifyCategory, on_delete=models.CASCADE, related_name='notifies',
                                 verbose_name='Категория')
    event = models.IntegerField(choices=settings.CHOICES_NOTIFY_EVENT, blank=True, null=True, verbose_name='Событие')
    files = models.ManyToManyField(NotifyFile, verbose_name='Файлы')

    is_read = models.BooleanField(default=False, verbose_name='Прочитано')
    data_json = models.TextField(blank=True, null=True, verbose_name='Данные пуш-уведомления (JSON)')

    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')
    send_at = models.DateTimeField(blank=True, null=True, verbose_name='Время начала отправки')
    sent_at = models.DateTimeField(blank=True, null=True, verbose_name='Дата отправки')

    def __str__(self):
        return self.subject

    def get_sender(self):
        pass

    def _send(self):
        if self.user:
            self.email = self.user.email
            self.phone = self.user.phone
            self.telegram_chat_id = self.user.telegram_chat_id
            self.viber_chat_id = self.user.viber_chat_id
        elif self.email:
            self.email = self.email

        if self.phone is not None:
            self.phone = re.sub("[^0-9]", "", self.phone)

        if self.type == TYPE.EMAIL:
            self.send_email()
        if self.type == TYPE.SMS:
            self.send_sms()
        if self.type == TYPE.PUSH:
            self.send_push()
        if self.type == TYPE.TELEGRAM:
            self.send_telegram()
        if self.type == TYPE.VIBER:
            self.send_viber()

        self.save()

    def send_email(self):
        account = self._get_valid_smtp_account()

        try:
            body = self._render_body(account.sender, self.category.template)
            server = SMTP_SSL(account.host, account.port) if account.is_use_ssl else SMTP(account.host, account.port)
            server.ehlo()
            if account.is_use_tls:
                server.starttls()
            server.login(account.username, account.password)
            server.sendmail(account.sender, [self.email], body.as_string())
            server.close()
            self.state = STATE.DELIVERED
            self.sent_at = now()
        except Exception as e:  # noqa
            self.state = STATE.REJECTED
            self.to_log(str(e))

    def send_sms(self):
        config = NotifyConfig.get_solo()

        if not config.is_sms_enabled:
            self.state = STATE.DISABLED
            return

        try:
            msg = str(self.text.replace(' ', '+'))
            url = '{url}?msg={text}&to={to}&api_id={api_id}&from={from_text}&json=1'.format(
                url=config.sms_url,
                api_id=config.sms_api_id,
                from_text=config.sms_from,
                to=self.phone,
                text=msg,
            )
            response = requests.get(url)
            response_dict = response.json()
            if response_dict.get('status'):
                self.state = STATE.DELIVERED
                self.sent_at = now()
            else:
                self.state = STATE.REJECTED
        except Exception as e:  # noqa
            self.state = STATE.REJECTED
            self.to_log(str(e))

    def send_push(self):
        config = NotifyConfig.get_solo()

        if not config.is_push_enabled:
            self.state = STATE.DISABLED
            return

        if self.user is None:
            self.state = STATE.REJECTED
            self.save()
            return

        devices = NotifyDevice.objects.filter(active=True, user=self.user).distinct('device_id')

        data_json = {}
        if self.data_json is not None:
            data_json = json.loads(self.data_json)
        data_json.update({
            'notify_id': self.id,
        })

        try:
            devices.send_message(title=self.subject, body=self.text, data=data_json)
            self.state = STATE.DELIVERED
            self.sent_at = now()
        except Exception as e:  # noqa
            self.state = STATE.REJECTED
            self.to_log(str(e))

    def send_telegram(self):
        import telegram
        config = NotifyConfig.get_solo()
        bot = telegram.Bot(token=config.telegram_api_key)

        if not config.is_telegram_enabled:
            self.state = STATE.DISABLED
            return

        try:
            result = False
            for chunk in chunks(str(self.text), 4096):
                result = bot.sendMessage(chat_id=self.telegram_chat_id, text=chunk, disable_web_page_preview=True)
            if result:
                self.state = STATE.DELIVERED
                self.sent_at = now()
            else:
                self.state = STATE.REJECTED
                self.to_log('REJECTED WITH DATA, please test it.')
        except Exception as e:  # noqa
            self.state = STATE.REJECTED
            self.to_log(str(e))

    @staticmethod
    def send(event, context, user=None, email=None, phone=None, files=None, data_json=None, notify_templates=None,
             viber_chat_id=None):
        from user.models import User

        if user:
            email = user.email if not email else email
            phone = user.phone if not phone else phone

        user_want_message_check = None
        if hasattr(settings, 'NOTIFY_USER_WANT_MESSAGE_CHECK') and settings.NOTIFY_USER_WANT_MESSAGE_CHECK is not None:
            user_want_message_check = import_string(settings.NOTIFY_USER_WANT_MESSAGE_CHECK)

        # Для выбора шаблонов в action'е
        if notify_templates:
            templates = NotifyTemplate.objects.filter(id__in=notify_templates, event=event, is_active=True)
        else:
            templates = NotifyTemplate.objects.filter(event=event, is_active=True)

        if files is None:
            files = []

        data_json = json.dumps(data_json) if data_json is not None else None

        file_instances = []
        for f in files:
            instance = NotifyFile.objects.create(file=f)
            file_instances.append(instance)

        notification_count = 0
        for template in templates:
            # Формируем список основных и дополнительных получателей письма
            # Первого получателя заполняем из шаблона, его данные могут быть пустыми - чтобы можно было через
            # код отправить уведомление
            if template.user or template.email or template.phone or template.viber_chat_id:
                recievers = [{
                    'user': template.user,
                    'email': template.email,
                    'phone': template.phone,
                    'viber_chat_id': template.viber_chat_id,
                }]
            elif user or phone or email or viber_chat_id:
                recievers = [{
                    'user': user,
                    'email': email,
                    'phone': phone,
                    'viber_chat_id': viber_chat_id,
                }]
            else:
                recievers = []

            for user_list in template.user_lists.all():
                for participant in user_list.participants.all():
                    # Если у участника из дополнительного списка ничего не заполнено - не отправляем уведомление
                    if participant.user is not None:
                        recievers.append({
                            'user': participant.user,
                            'email': participant.email,
                            'phone': participant.phone,
                            'viber_chat_id': participant.viber_chat_id,
                        })
                    elif participant.email is not None:
                        recievers.append({
                            'user': participant.user,
                            'email': participant.email,
                            'phone': participant.phone,
                            'viber_chat_id': participant.viber_chat_id,
                        })

                # Если в списке отмечены группы
                if user_list.user_groups and not user_list.mail_to_all:
                    group_users = User.objects.filter(groups__in=list(user_list.user_groups.all()))
                    for user in group_users:
                        recievers.append(
                            {
                                'user': user,
                                'email': user.email,
                                'phone': user.phone,
                                'viber_chat_id': user.viber_chat_id,
                            }
                        )
                    # Убираем дубликаты пользователей
                    recievers = list({v['email']: v for v in recievers}.values())

                # Если в списке пользователей отмечены все пользователи
                if user_list.mail_to_all:
                    recievers = []
                    users = User.objects.all()
                    for user in users:
                        recievers.append(
                            {
                                'user': user,
                                'email': user.email,
                                'phone': user.phone,
                                'viber_chat_id': user.viber_chat_id,
                            }
                        )
            for reciever in recievers:
                template_user = reciever['user']
                template_email = reciever['email']
                template_phone = reciever['phone']
                template_viber_chat_id = reciever['viber_chat_id']
                # Если в шаблоне не указан получатель, то получатель тот, кого передали в коде
                if template_user is None and template_email is None and template_phone is None and template_viber_chat_id is None:
                    template_user = user
                    template_email = email
                    template_phone = phone
                    template_viber_chat_id = viber_chat_id
                    # Если пользователь заполнен, то перезаписываем поля емейла и номера телефона,
                    # даже если они были переданы.
                    if user is not None:
                        template_email = user.email
                        template_phone = user.phone
                        template_viber_chat_id = user.viber_chat_id
                # Проверка, хочет ли получить сообщение
                if user_want_message_check is not None and not \
                        user_want_message_check(event, template.type, template_user):  # noqa
                    continue

                local_context = context.copy()

                # Добавляем пользователя в контекст, если его там не передали
                if template_user is not None:
                    if local_context is not None:
                        local_context.update({
                            'user': template_user,
                        })
                    else:
                        local_context = {
                            'user': template_user,
                        }
                local_context['event_id'] = event

                instance = Notify.objects.create(
                    subject=template.render_subject(local_context),
                    text=template.render_text(local_context),
                    html=template.render_html(local_context),
                    user=template_user,
                    email=template_email,
                    phone=template_phone if template_phone is not None else "",
                    viber_chat_id=template_viber_chat_id if template_viber_chat_id is not None else "",
                    type=template.type,
                    event=template.event,
                    category=template.category,
                    data_json=data_json,
                    send_at=template.send_at,
                )
                file_instance = instance.files
                for f in file_instances:
                    file_instance.add(f)
                instance.save()
                notification_count += 1

        return notification_count

    def _render_body(self, mail_from, layout):
        msg = MIMEMultipart('alternative')

        msg['Subject'] = self.subject
        msg['From'] = mail_from
        msg['To'] = self.email

        text = MIMEText(self.text, 'plain')
        msg.attach(text)

        if self.html:
            template = Template(layout)
            context = Context({'text': mark_safe(self.html)})
            html = MIMEText(mark_safe(template.render(context)), 'html')
            msg.attach(html)

        for fl in self.files.all():
            with fl.file.open(mode='rb') as f:
                part = MIMEApplication(
                    f.read(),
                    Name=fl.file.name.split('/')[-1]
                )

            part['Content-Disposition'] = 'attachment; filename="%s"' % fl.file.name.split('/')[-1]

            msg.attach(part)

        return msg

    def get_format_state(self):
        if self.state == STATE.WAIT:
            return format_html('<span style="color:orange;">В ожидании</span>')
        elif self.state == STATE.DELIVERED:
            return format_html('<span style="color:green;">Отправлено</span>')
        elif self.state == STATE.REJECTED:
            return format_html('<span style="color:red;">Отклонено</span>')
        elif self.state == STATE.DISABLED:
            return format_html('<span style="color:red;">Отправка запрещена</span>')
        return format_html('<span style="color:black;">Неизвестный статус</span>')

    get_format_state.short_description = 'Статус'

    def to_log(self, error_text):
        log = NotifyErrorLog(notify=self, error=error_text)
        log.save()

    def send_viber(self):
        config = NotifyConfig.get_solo()
        viber = Api(BotConfiguration(
            name=config.viber_bot_name,
            avatar='',
            auth_token=config.viber_api_key
        ))
        if not config.is_viber_enabled:
            self.state = STATE.DISABLED
            return
        try:
            result = False
            # из шаблона получаем выбраные списки пользователей для рассылки
            for user_lists_in_template in NotifyTemplate.objects.all().prefetch_related('user_lists'):
                # проверяем добавлены ли списки пользователей для рассылки в шаблон
                if user_lists_in_template.user_lists.all().exists():
                    # перебираем списки пользователей для рассылки
                    for participants_in_user_lists in user_lists_in_template.user_lists.all():
                        # перебираем участников списка пользователей
                        for participant in participants_in_user_lists.participants.all():
                            result = viber.send_messages(to=participant.user.viber_chat_id,
                                                         messages=[TextMessage(text=self.text)])
                # если не добавлены, сообщение приходит пользователю(получателю) или тому, кто указан в коде
                else:
                    result = viber.send_messages(to=self.viber_chat_id, messages=[TextMessage(text=self.text)])
            if result:
                self.state = STATE.DELIVERED
                self.sent_at = now()
            else:
                self.state = STATE.REJECTED
                self.to_log('REJECTED WITH DATA, please test it.')
        except Exception as e:  # noqa
            self.state = STATE.REJECTED
            self.to_log(str(e))

    def _get_valid_smtp_account(self) -> Optional[SMTPAccount]:
        config = NotifyConfig.get_solo()
        if not config.is_email_enabled:
            self.state = STATE.DISABLED
            return

        account = SMTPAccount.get_free_smtp()
        self.sender_email = account.sender
        if account is None or not account.is_active:
            return

        return account

    class Meta:
        verbose_name = 'Уведомление'
        verbose_name_plural = 'Уведомления'



from django.utils import timezone
from app.celery import app as celery_app
from .models.choices import STATE
from .models.config import NotifyConfig
from .models.notify import Notify


@celery_app.task
def send_notifications():
    notifies = Notify.objects.filter(state__in=[STATE.WAIT])

    for notify in notifies.iterator():
        if notify.state == STATE.WAIT:
            if notify.send_at is None:
                notify._send()
            else:
                if timezone.now() > notify.send_at:
                    notify._send()
    return


celery_app.conf.beat_schedule.update({
    'periodic_task': {
        'task': 'garpix_notify.tasks.send_notifications',
        'schedule': NotifyConfig.get_solo().periodic,
    },
})
celery_app.conf.timezone = 'UTC'

# celery_app.add_periodic_task(NotifyConfig.get_solo().periodic, send_notifications.s(), name='periodic_task') #2 способ



from django.contrib import admin
from ..models.smtp import SMTPAccount


def clear_limits(modeladmin, request, queryset):
    for server in queryset.all():
        server.clear()


clear_limits.short_description = 'Обнулить лимиты'


@admin.register(SMTPAccount)
class SMTPAccountAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'is_worked_now')
    readonly_fields = ('email_hour_used_times', 'email_hour_used_date', 'email_day_used_times', 'email_day_used_date')
    actions = [
        clear_limits
    ]



from django.db import models
from ..utils import get_file_path


class NotifyFile(models.Model):
    file = models.FileField(upload_to=get_file_path, max_length=1000, verbose_name='Файл')
    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')

    def __str__(self):
        return str(self.file.url) if self.file else '-'

    class Meta:
        verbose_name = 'Файл'
        verbose_name_plural = 'Файлы'



from django.contrib import admin
from fcm_django.models import FCMDevice
from ..models.fcm import NotifyDevice

admin.site.unregister(FCMDevice)


@admin.register(NotifyDevice)
class NotifyDeviceAdmin(admin.ModelAdmin):
    pass



import datetime
import uuslug


def get_file_path(instance, filename):
    """
    Формирует путь файла относительно года и месяца, чтобы множество файлов не скапливались на одном уровне.
    """
    ext = filename.split('.')[-1]
    today = datetime.date.today()
    filename = f'{uuslug.slugify(".".join(filename.split(".")[:-1]))}.{ext}'
    return f'uploads/{today.year}/{today.month}/{filename}'



from .category import NotifyCategoryAdmin
from .config import NotifyConfigAdmin
from .notify import NotifyAdmin
from .smtp import SMTPAccountAdmin
from .fcm import NotifyDeviceAdmin
from .template import NotifyTemplateAdmin
from .file import NotifyFileAdmin
from .user_list import NotifyUserListAdmin



from django.db import models
from .user_list import NotifyUserList
from django.conf import settings
from ..mixins import UserNotifyMixin


class NotifyUserListParticipant(UserNotifyMixin):
    user_list = models.ForeignKey(NotifyUserList, on_delete=models.CASCADE, related_name='participants', verbose_name='Список пользователей для рассылки')
    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL,
                             related_name='user_lists', verbose_name='Пользователь (получатель)')
    email = models.EmailField(max_length=255, blank=True, null=True, verbose_name='Email Получателя',
                              help_text='Используется только в случае отсутствия указанного пользователя')

    def __str__(self):
        return f'Участник рассылки {self.pk}'

    class Meta:
        verbose_name = 'Участник списка пользователей'
        verbose_name_plural = 'Участники списка пользователей'



from django.contrib import admin
from solo.admin import SingletonModelAdmin
from ..models.config import NotifyConfig


@admin.register(NotifyConfig)
class NotifyConfigAdmin(SingletonModelAdmin):
    fieldsets = (
        (None, {
            'fields': ('periodic', 'is_email_enabled', 'is_sms_enabled', 'is_push_enabled', 'is_telegram_enabled',
                       'is_viber_enabled')
        }),
        ('Почта', {
            'fields': ('email_max_day_limit', 'email_max_hour_limit')
        }),
        ('СМС', {
            'fields': ('sms_url', 'sms_api_id', 'sms_from')
        }),
        ('Telegram', {
            'fields': ('telegram_api_key',)
        }),
        ('Viber', {
            'fields': ('viber_api_key', 'viber_bot_name', 'viber_success_added_text',
                       'viber_failed_added_text', 'viber_text_for_new_sub', 'viber_welcome_text')
        })
    )



from django.contrib import admin
from ..models.user_list import NotifyUserList
from ..models.user_list_participant import NotifyUserListParticipant


class NotifyUserListParticipantInline(admin.TabularInline):
    model = NotifyUserListParticipant
    raw_id_fields = ('user', )
    fields = ('user', 'email')


@admin.register(NotifyUserList)
class NotifyUserListAdmin(admin.ModelAdmin):
    list_display = ('title',)
    inlines = [
        NotifyUserListParticipantInline
    ]
    readonly_fields = ('created_at',)
    search_fields = ('title',)



class TYPE:
    EMAIL = 0
    SMS = 1
    PUSH = 2
    TELEGRAM = 3
    VIBER = 4

    CHOICES = (
        (EMAIL, 'E-mail'),
        (SMS, 'SMS'),
        (PUSH, 'Push'),
        (TELEGRAM, 'Telegram'),
        (VIBER, 'Viber'),
    )


class STATE:
    DISABLED = -2
    REJECTED = -1
    WAIT = 0
    DELIVERED = 1

    CHOICES = (
        (DELIVERED, 'Доставлено'),
        (REJECTED, 'Отклонено'),
        (WAIT, 'В ожидании'),
        (DISABLED, 'Не отправлено (отправка запрещена настройками)'),
    )



from django.test import TestCase
from .models import NotifyTemplate
from .models import NotifyCategory
from .models import Notify
from .models import NotifyUserList
from .models import NotifyUserListParticipant
from .models.choices import TYPE
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group


User = get_user_model()


class PreBuildTestCase(TestCase):
    def setUp(self):
        # нулевой евент, только для теста
        self.PASS_TEST_EVENT = 0
        # тестовый пользователь
        self.data_user = {
            'username': 'test',
            'email': 'test@garpix.com',
            'password': 'BlaBla123',
            'first_name': 'Ivan',
            'last_name': 'Ivanov',
        }
        # data
        self.data_category = {
            'title': 'Основная категория',
            'template': '<div>{{text}}</div>',
        }
        self.data_template_email = {
            'title': 'Тестовый темплейт',
            'subject': 'Тестовый темплейт {{user.email}}',
            'text': 'Контент текстовый {{user.email}} - {{sometext}}',
            'html': 'Контент HTML {{user.email}} - {{sometext}}',
            'type': TYPE.EMAIL,
            'event': self.PASS_TEST_EVENT,
        }
        self.sometext = 'bla bla'
        self.data_compiled_email = {
            'subject': f'Тестовый темплейт {self.data_user["email"]}',
            'text': f'Контент текстовый {self.data_user["email"]} - {self.sometext}',
            'html': f'Контент HTML {self.data_user["email"]} - {self.sometext}',
            'type': TYPE.EMAIL,
            'event': self.PASS_TEST_EVENT,
        }
        super().setUp()

    def test_notify_email_positive(self):
        # Создание пользователя
        user = User.objects.create_user(**self.data_user)
        # Создание категории
        category = NotifyCategory.objects.create(**self.data_category)
        category = NotifyCategory.objects.get(pk=category.pk)
        self.assertEqual(category.title, self.data_category['title'])
        self.assertEqual(category.template, self.data_category['template'])
        # Создание шаблона
        template_email = NotifyTemplate.objects.create(category=category, **self.data_template_email)
        template_email = NotifyTemplate.objects.get(pk=template_email.pk)
        self.assertEqual(template_email.title, self.data_template_email['title'])
        self.assertEqual(template_email.subject, self.data_template_email['subject'])
        self.assertEqual(template_email.text, self.data_template_email['text'])
        self.assertEqual(template_email.html, self.data_template_email['html'])
        self.assertEqual(template_email.type, self.data_template_email['type'])
        self.assertEqual(template_email.event, self.data_template_email['event'])
        self.assertEqual(template_email.category.id, category.id)
        # Создание пробного письма
        Notify.send(self.PASS_TEST_EVENT, {
            'sometext': self.sometext,
            'user': user,
        }, user=user)
        self.assertEqual(Notify.objects.all().count(), 1)
        notify = Notify.objects.all().first()
        self.assertEqual(notify.subject, self.data_compiled_email['subject'])
        self.assertEqual(notify.text, self.data_compiled_email['text'])
        self.assertEqual(notify.html, self.data_compiled_email['html'])
        self.assertEqual(notify.type, self.data_compiled_email['type'])
        self.assertEqual(notify.event, self.data_compiled_email['event'])

    def test_notify_email_user_list(self):
        # Создание пользователя
        user = User.objects.create_user(**self.data_user)
        # Создание категории
        category = NotifyCategory.objects.create(**self.data_category)
        category = NotifyCategory.objects.get(pk=category.pk)
        self.assertEqual(category.title, self.data_category['title'])
        self.assertEqual(category.template, self.data_category['template'])
        # Создание списка
        user_list = NotifyUserList.objects.create(title='Админы')
        group = Group.objects.create(name="Админы клиенты")
        user_list.user_groups.add(group)

        user_list_participant1 = NotifyUserListParticipant.objects.create(  # noqa
            user_list=user_list,
            email='test2@garpix.com',
        )
        user_list_participant2 = NotifyUserListParticipant.objects.create(  # noqa
            user_list=user_list,
        )
        user_list_participant3 = NotifyUserListParticipant.objects.create(  # noqa
            user_list=user_list,
            email='test3@garpix.com',
        )

        # Создание шаблона
        template_email = NotifyTemplate.objects.create(category=category, **self.data_template_email)
        template_email.user_lists.add(user_list)
        template_email = NotifyTemplate.objects.get(pk=template_email.pk)

        self.assertEqual(template_email.title, self.data_template_email['title'])
        self.assertEqual(template_email.subject, self.data_template_email['subject'])
        self.assertEqual(template_email.text, self.data_template_email['text'])
        self.assertEqual(template_email.html, self.data_template_email['html'])
        self.assertEqual(template_email.type, self.data_template_email['type'])
        self.assertEqual(template_email.event, self.data_template_email['event'])
        self.assertEqual(template_email.category.id, category.id)

        # Создание пробного письма
        Notify.send(self.PASS_TEST_EVENT, {
            'sometext': self.sometext,
            'user': user,
        }, user=user)

        self.assertEqual(Notify.objects.all().count(), 3)

        # notify 1
        notify = Notify.objects.get(user=user)
        self.assertEqual(notify.subject, self.data_compiled_email['subject'])
        self.assertEqual(notify.text, self.data_compiled_email['text'])
        self.assertEqual(notify.html, self.data_compiled_email['html'])
        self.assertEqual(notify.type, self.data_compiled_email['type'])
        self.assertEqual(notify.event, self.data_compiled_email['event'])
        self.assertEqual(notify.email, 'test@garpix.com')
        # notify 2
        notify = Notify.objects.get(email='test2@garpix.com')
        self.assertEqual(notify.subject, self.data_compiled_email['subject'])
        self.assertEqual(notify.text, self.data_compiled_email['text'])
        self.assertEqual(notify.html, self.data_compiled_email['html'])
        self.assertEqual(notify.type, self.data_compiled_email['type'])
        self.assertEqual(notify.event, self.data_compiled_email['event'])
        self.assertEqual(notify.email, 'test2@garpix.com')
        # notify 3
        notify = Notify.objects.get(email='test3@garpix.com')
        self.assertEqual(notify.subject, self.data_compiled_email['subject'])
        self.assertEqual(notify.text, self.data_compiled_email['text'])
        self.assertEqual(notify.html, self.data_compiled_email['html'])
        self.assertEqual(notify.type, self.data_compiled_email['type'])
        self.assertEqual(notify.event, self.data_compiled_email['event'])
        self.assertEqual(notify.email, 'test3@garpix.com')

    def test_notify_viber(self):
        self.data_template_viber = {
            'title': 'Тестовый темплейт',
            'subject': 'Тестовый темплейт {{user.viber_chat_id}}',
            'text': 'Контент текстовый {{user.viber_chat_id}} - {{sometext}}',
            'html': 'Контент HTML {{user.viber_chat_id}} - {{sometext}}',
            'type': TYPE.VIBER,
            'event': self.PASS_TEST_EVENT,
        }
        self.data_viber_user = {
            'username': 'test',
            'viber_secret_key': '111',
            'viber_chat_id': 'm4FsaRu5kBi8HzSAC0liFQ==',
            'password': 'BlaBla123',
            'first_name': 'Ivan',
            'last_name': 'Ivanov',
        }
        self.data_compiled_viber = {
            'subject': f'Тестовый темплейт {self.data_viber_user["viber_chat_id"]}',
            'text': f'Контент текстовый {self.data_viber_user["viber_chat_id"]} - {self.sometext}',
            'html': f'Контент HTML {self.data_viber_user["viber_chat_id"]} - {self.sometext}',
            'type': TYPE.VIBER,
            'event': self.PASS_TEST_EVENT,
        }
        # Создание пользователя
        user = User.objects.create_user(**self.data_viber_user)
        # Создание категории
        category = NotifyCategory.objects.create(**self.data_category)
        category = NotifyCategory.objects.get(pk=category.pk)
        self.assertEqual(category.title, self.data_category['title'])
        self.assertEqual(category.template, self.data_category['template'])
        # Создание списка
        user_list = NotifyUserList.objects.create(title='viber_user_list')
        user_list_participant1 = NotifyUserListParticipant.objects.create(  # noqa
            user_list=user_list,
        )
        template_viber = NotifyTemplate.objects.create(category=category, **self.data_template_viber)
        template_viber.user_lists.add(user_list)
        template_viber = NotifyTemplate.objects.get(pk=template_viber.pk)
        self.assertEqual(template_viber.title, self.data_template_viber['title'])
        self.assertEqual(template_viber.subject, self.data_template_viber['subject'])
        self.assertEqual(template_viber.text, self.data_template_viber['text'])
        self.assertEqual(template_viber.html, self.data_template_viber['html'])
        self.assertEqual(template_viber.type, self.data_template_viber['type'])
        self.assertEqual(template_viber.event, self.data_template_viber['event'])
        self.assertEqual(template_viber.category.id, category.id)

        Notify.send(self.PASS_TEST_EVENT, {
            'sometext': self.sometext,
            'user': user,
        }, user=user)
        self.assertEqual(Notify.objects.all().count(), 1)

        # notify 1
        notify = Notify.objects.get(user=user)
        self.assertEqual(notify.subject, self.data_compiled_viber['subject'])
        self.assertEqual(notify.text, self.data_compiled_viber['text'])
        self.assertEqual(notify.html, self.data_compiled_viber['html'])
        self.assertEqual(notify.type, self.data_compiled_viber['type'])
        self.assertEqual(notify.event, self.data_compiled_viber['event'])
        self.assertEqual(notify.user.viber_chat_id, 'm4FsaRu5kBi8HzSAC0liFQ==')



from django.db import models


class UserNotifyMixin(models.Model):
    phone = models.CharField(max_length=30, blank=True, default='', verbose_name='Телефон')
    telegram_chat_id = models.CharField(max_length=200, blank=True, default='',
                                        verbose_name='Telegram ID пользователя/чата')
    viber_chat_id = models.CharField(max_length=200, blank=True, default='',
                                     verbose_name='Viber ID пользователя/чата')
    viber_secret_key = models.CharField(max_length=200, blank=True, default='',
                                        verbose_name='Ключ подключения Viber')

    class Meta:
        abstract = True



from django.contrib import admin
from ..models.notify import Notify
from .log import NotifyErrorLogInline


class FileInline(admin.TabularInline):
    model = Notify.files.through
    verbose_name = 'Файл'
    verbose_name_plural = 'Файлы'
    raw_id_fields = ('notifyfile',)
    extra = 0


@admin.register(Notify)
class NotifyAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'type', 'event', 'user', 'email', 'phone', 'get_format_state', 'created_at', 'send_at',
                    'sent_at')
    inlines = [
        FileInline,
        NotifyErrorLogInline
    ]
    readonly_fields = ('created_at',)
    exclude = ('files',)
    list_filter = ('event', 'state', 'type', 'category')
    search_fields = ('subject', 'text')
    raw_id_fields = ('user',)



from django.db import models
from django.conf import settings
from django.template import Template, Context

from .choices import TYPE
from .category import NotifyCategory
from .user_list import NotifyUserList

from ckeditor_uploader.fields import RichTextUploadingField
from ..mixins import UserNotifyMixin


class NotifyTemplate(UserNotifyMixin):
    title = models.CharField(max_length=255, verbose_name='Название для админа')
    subject = models.CharField(max_length=255, default='', blank=True, verbose_name='Заголовок')
    text = models.TextField(verbose_name='Текст')
    html = RichTextUploadingField(verbose_name='HTML')

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, blank=True, null=True, verbose_name='Пользователь (получатель)')
    email = models.EmailField(max_length=255, blank=True, null=True, verbose_name='Email получатель', help_text='Используется только в случае отсутствия указанного пользователя')

    type = models.IntegerField(choices=TYPE.CHOICES, verbose_name='Тип')
    category = models.ForeignKey(NotifyCategory, on_delete=models.CASCADE, related_name='templates', verbose_name='Категория')
    event = models.IntegerField(choices=settings.CHOICES_NOTIFY_EVENT, blank=True, null=True, verbose_name='Событие')

    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')
    is_active = models.BooleanField(default=True, verbose_name='Активный')
    send_at = models.DateTimeField(blank=True, null=True, verbose_name='Время начала отправки')

    user_lists = models.ManyToManyField(NotifyUserList, blank=True, verbose_name='Списки пользователей, которые получат копию уведомления')

    def render_subject(self, ct):
        template = Template(self.subject)
        context = Context(ct)
        return template.render(context)

    def render_text(self, ct):
        template = Template(self.text)
        context = Context(ct)
        return template.render(context)

    def render_html(self, ct):
        template = Template(self.html)
        context = Context(ct)
        return template.render(context)

    def get_event_data(self):
        return settings.NOTIFY_EVENTS[self.event]

    def get_context_description(self):
        try:
            text = self.get_event_data()['context_description'] if self.event else '--'
        except:  # noqa
            text = '--'
        return text
    get_context_description.short_description = 'Описание контекста шаблона'

    def get_event_description(self):
        try:
            text = self.get_event_data()['event_description'] if self.event else '--'
        except:  # noqa
            text = '--'
        return text
    get_event_description.short_description = 'Описание события'

    def get_test_data(self):
        try:
            data = self.get_event_data()['test_data'] if self.event else {}
        except:  # noqa
            data = {}
        return data

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = 'Шаблон'
        verbose_name_plural = 'Шаблоны'
